from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

from lxml import etree
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig

from vunnel.providers.mariner.model import Definition, RpminfoObject, RpminfoState, RpminfoTest
from vunnel.utils import http
from vunnel.utils.vulnerability import FixedIn, Vulnerability

if TYPE_CHECKING:
    import logging
    from collections.abc import Generator

    from vunnel.workspace import Workspace

LESS_THAN_OR_EQUAL_TO = "less than or equal"
LESS_THAN = "less than"
GREATER_THAN = "greater than"

IGNORED_PATCHABLE_VALUES = ["Not Applicable"]


class MarinerXmlFile:
    def __init__(self, oval_file_path: str, logger: logging.Logger):
        parser_config = ParserConfig(
            fail_on_converter_warnings=False,
            fail_on_unknown_attributes=False,
            fail_on_unknown_properties=False,
        )
        xml_parser = XmlParser(config=parser_config)
        # S320 disable explanation: the mariner linux vulnerability feed is not untrusted xml
        root = etree.parse(oval_file_path)  # noqa: S320
        nsmap = etree.XPath("/*")(root)[0].nsmap
        default = nsmap[None]
        nsmap["default"] = default
        del nsmap[None]
        self.logger = logger

        # Go element by element to have flexible error handling
        self.definitions = []
        for def_element in etree.XPath("//default:definition", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(def_element)
                definition = xml_parser.from_string(element_bytes.decode(), Definition)
                self.definitions.append(definition)
            except Exception as ex:
                self.logger.warning(f"skipping definition element in {oval_file_path} due to {ex}")

        self.tests_by_id = {}
        for test_element in etree.XPath("//linux-def:rpminfo_test", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(test_element)
                test_obj = xml_parser.from_string(element_bytes.decode(), RpminfoTest)
                self.tests_by_id[test_obj.id] = test_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_test element in {oval_file_path} due to {ex}")
        self.objects_by_id = {}
        for obj_element in etree.XPath("//linux-def:rpminfo_object", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(obj_element)
                obj = xml_parser.from_string(element_bytes.decode(), RpminfoObject)
                self.objects_by_id[obj.id] = obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")

        self.states_by_id = {}
        for state_element in etree.XPath("//linux-def:rpminfo_state", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(state_element)
                state_obj = xml_parser.from_string(element_bytes.decode(), RpminfoState)
                self.states_by_id[state_obj.id] = state_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")
        self.mariner_version = oval_file_path.split("-")[-2]

    def name_and_version(self, test_id: str) -> tuple[str | None, str | None]:
        test = self.tests_by_id.get(test_id, None)
        if test is None or test.object_value is None or test.state is None:
            return (None, None)
        obj = self.objects_by_id.get(test.object_value.object_ref, None)
        if obj is None:
            return (None, None)
        state = self.states_by_id.get(test.state.state_ref, None)
        if state is None or state.evr is None:
            return (None, None)
        return (obj.name, state.evr.value)

    def namespace_name(self) -> str:
        return f"mariner:{self.mariner_version}"

    def get_tests(self, definition: Definition) -> list[RpminfoTest]:
        tests = []
        if definition and definition.criteria and definition.criteria.criterion:
            for criterion in definition.criteria.criterion:
                test = self.tests_by_id.get(criterion.test_ref, None)
                if test:
                    tests.append(test)
        return tests

    def get_states(self, tests: list[RpminfoTest]) -> list[RpminfoState]:
        states = []
        for test in tests:
            if test and test.state and test.state.state_ref:
                state = self.states_by_id.get(test.state.state_ref, None)
                if state:
                    states.append(state)
        return states

    def get_objects(self, tests: list[RpminfoTest]) -> list[RpminfoObject]:
        objects = []
        for test in tests:
            if test and test.object_value and test.object_value.object_ref:
                obj = self.objects_by_id.get(test.object_value.object_ref, None)
                if obj:
                    objects.append(obj)
        return objects

    def make_fixed_in(self, definition: Definition) -> FixedIn | None:
        tests = self.get_tests(definition)
        states = self.get_states(tests)
        objects = self.get_objects(tests)

        if not states or not objects:
            return None

        name = objects[0].name
        if not name:
            return None

        fixed_version = None

        vulnerability_range = []
        for state in states:
            if state.evr and state.evr.value:
                if state.evr.operation == LESS_THAN:
                    vulnerability_range.append(f"< {state.evr.value}")
                    # if vulnerability has an upper bound (< as opposed to <=),
                    # then assume the upper bound is the fixed version
                    fixed_version = state.evr.value
                elif state.evr.operation == GREATER_THAN:
                    vulnerability_range.append(f"> {state.evr.value}")
                elif state.evr.operation == LESS_THAN_OR_EQUAL_TO:
                    vulnerability_range.append(f"<= {state.evr.value}")

        if not vulnerability_range:
            return None

        # make output deterministic. Reverse so that
        # output reads like >1.2.3 <1.3.0 instead of the reverse.
        vulnerability_range.sort(reverse=True)

        if not fixed_version:
            fixed_version = "None"  # a required string in JSON schema

        vulnerability_range_str = ", ".join(vulnerability_range)

        return FixedIn(
            Name=name,
            NamespaceName=self.namespace_name(),
            VersionFormat="rpm",
            Version=fixed_version,
            VulnerableRange=vulnerability_range_str,
            Module=None,
            VendorAdvisory=None,
        )

    def vulnerability_id(self, definition: Definition) -> str | None:
        if definition.metadata is None or definition.metadata.reference is None or definition.metadata.reference.ref_id is None:
            return None
        return definition.metadata.reference.ref_id

    def description(self, definition: Definition) -> str | None:
        if not definition.metadata:
            return None
        return definition.metadata.description

    def vulnerabilities(self) -> Generator[Vulnerability, None, None]:
        for d in self.definitions:
            if d.metadata is None or d.metadata.severity is None:
                self.logger.warning("skipping definition because severity could not be found")
                continue
            if d.metadata and d.metadata.patchable and d.metadata.patchable in IGNORED_PATCHABLE_VALUES:
                continue
            link = ""
            if d.metadata.reference and d.metadata.reference.ref_url:
                link = d.metadata.reference.ref_url
            fixed_in = self.make_fixed_in(d)
            if not fixed_in:
                continue
            vulnerability_id = self.vulnerability_id(d)
            if not vulnerability_id:
                continue
            yield Vulnerability(
                Name=vulnerability_id,  # intentional; legacy API uses Name as field for vulnerability ID.
                NamespaceName=self.namespace_name(),
                Description=self.description(d) or "",
                Severity=d.metadata.severity,
                Link=link,
                CVSS=[],
                FixedIn=[fixed_in],
                Metadata={},  # Ideally, there's no metadata here.
            )


MARINER_URL_BASE = "https://raw.githubusercontent.com/microsoft/CBL-MarinerVulnerabilityData/main/{}"
MARINER_URL_FILENAME = "cbl-mariner-{}-oval.xml"
AL3_URL = "https://raw.githubusercontent.com/microsoft/AzureLinuxVulnerabilityData/main/azurelinux-3.0-oval.xml"

VERSION_TO_URL = {
    "1.0": MARINER_URL_BASE.format(MARINER_URL_FILENAME.format("1.0")),
    "2.0": MARINER_URL_BASE.format(MARINER_URL_FILENAME.format("2.0")),
    "3.0": AL3_URL,
}

VERSION_TO_FILENAME = {
    "1.0": MARINER_URL_FILENAME.format("1.0"),
    "2.0": MARINER_URL_FILENAME.format("2.0"),
    "3.0": "azurelinux-3.0-oval.xml",
}


class Parser:
    def __init__(self, workspace: Workspace, download_timeout: int, allow_versions: list[Any], logger: logging.Logger):
        self.workspace = workspace
        self.download_timeout = download_timeout
        self.allow_versions = allow_versions
        self._urls: set[str] = set()
        self.logger = logger

    def _download(self) -> list[str]:
        return [self._download_version(v) for v in self.allow_versions]

    def _download_version(self, version: str) -> str:
        filename = VERSION_TO_FILENAME[version]
        if not filename:
            raise Exception(f"mariner/azurelinux provider misconfigured: no filename for version {version}")
        url = VERSION_TO_URL[version]
        if not url:
            raise Exception(f"mariner/azurelinux provider misconfigured: no URL for version {version}")
        r = http.get(url, self.logger, timeout=self.download_timeout)
        destination = os.path.join(self.workspace.input_path, filename)
        with open(destination, "wb") as writer:
            writer.write(r.content)
        self._urls.add(url)
        return destination

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def get(self) -> Generator[tuple[str, str, dict[str, dict[str, Any]]], None, None]:
        for oval_file_path in self._download():
            parsed_file = MarinerXmlFile(oval_file_path, self.logger)
            for v in parsed_file.vulnerabilities():
                yield v.NamespaceName, v.Name, v.to_payload()



from __future__ import annotations

import os.path
from dataclasses import dataclass

# Note: this metadata.json file currently is not allowed to have a breaking change
PROVIDER_WORKSPACE_STATE_SCHEMA_VERSION = "1.0.3"

PROVIDER_ARCHIVE_LISTING_SCHEMA_VERSION = "1.0.0"
MATCH_EXCLUSION_SCHEMA_VERSION = "1.0.0"
GITHUB_SECURITY_ADVISORY_SCHEMA_VERSION = "1.0.1"
MSRC_SCHEMA_VERSION = "1.0.0"
OS_SCHEMA_VERSION = "1.0.0"
NVD_SCHEMA_VERSION = "1.0.0"
OSV_SCHEMA_VERSION = "1.6.1"


@dataclass(frozen=True)
class Schema:
    version: str
    url: str

    @property
    def major_version(self) -> str:
        return self.version.split(".")[0]

    @property
    def name(self) -> str:
        name = self.url.removeprefix("https://raw.githubusercontent.com/anchore/vunnel/main/schema/")
        return os.path.dirname(name)


def ProviderListingSchema(version: str = PROVIDER_ARCHIVE_LISTING_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/provider-archive-listing/schema-{version}.json",
    )


def ProviderStateSchema(version: str = PROVIDER_WORKSPACE_STATE_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/provider-workspace-state/schema-{version}.json",
    )


def MatchExclusionSchema(version: str = MATCH_EXCLUSION_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/match-exclusion/schema-{version}.json",
    )


def GithubSecurityAdvisorySchema(version: str = GITHUB_SECURITY_ADVISORY_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/github-security-advisory/schema-{version}.json",
    )


def MSRCSchema(version: str = MSRC_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/msrc/schema-{version}.json",
    )


def OSSchema(version: str = OS_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/os/schema-{version}.json",
    )


def NVDSchema(version: str = NVD_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/nvd/schema-{version}.json",
    )


def OSVSchema(version: str = OSV_SCHEMA_VERSION) -> Schema:
    return Schema(
        version=version,
        url=f"https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/osv/schema-{version}.json",
    )



from __future__ import annotations

import bz2
import logging
import os
import re

from vunnel.utils import http, rpm
from vunnel.utils.oval_parser import Config, parse

# One time initialization of driver specific configuration
ol_config = Config()

# regexes
ol_config.tag_pattern = re.compile(r"\{http://oval.mitre.org/XMLSchema/.*\}(\w*)")
ol_config.ns_pattern = re.compile(r"(\{http://oval.mitre.org/XMLSchema/.*\})\w*")
ol_config.is_installed_pattern = re.compile(r"Oracle Linux (\d+).*is installed")
ol_config.pkg_version_pattern = re.compile(r"(.*) is earlier than (.*)")
ol_config.pkg_module_pattern = re.compile(r"Module (.*) is enabled")
ol_config.signed_with_pattern = re.compile(r"(.*) is signed with the (.*) key")
ol_config.platform_version_pattern = re.compile(r"Oracle Linux (\d+)")

# xpath queries
ol_config.title_xpath_query = "{0}metadata/{0}title"
ol_config.severity_xpath_query = "{0}metadata/{0}advisory/{0}severity"
ol_config.platform_xpath_query = "{0}metadata/{0}affected/{0}platform"
ol_config.date_issued_xpath_query = "{0}metadata/{0}advisory/{0}issued"
ol_config.date_updated_xpath_query = None  # oracle oval does not contain date updated
ol_config.description_xpath_query = "{0}metadata/{0}description"
ol_config.sa_ref_xpath_query = '{0}metadata/{0}reference[@source="elsa"]'
ol_config.cve_xpath_query = "{0}metadata/{0}advisory/{0}cve"
ol_config.criteria_xpath_query = "{0}criteria"
ol_config.criterion_xpath_query = ".//{0}criterion"

# maps
ol_config.severity_dict = {
    "n/a": "Negligible",
    "low": "Low",
    "moderate": "Medium",
    "important": "High",
    "critical": "Critical",
}

# string formats
ol_config.ns_format = "ol:{}"


class Parser:
    _url_ = "https://linux.oracle.com/security/oval/com.oracle.elsa-all.xml.bz2"
    _xml_file_ = "com.oracle.elsa-all.xml"

    def __init__(self, workspace, config=None, download_timeout=125, logger=None):
        self.config = config if config else ol_config
        self.download_timeout = download_timeout
        self.xml_file_path = os.path.join(workspace.input_path, self._xml_file_)

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    @property
    def urls(self):
        return [self._url_]

    def _download(self):
        try:
            self.logger.info(f"downloading ELSA from {self._url_}")
            r = http.get(self._url_, self.logger, stream=True, timeout=self.download_timeout)
            if r.status_code != 200:
                raise Exception(f"GET {self._url_} failed with HTTP error {r.status_code}")

            with open(self.xml_file_path, "wb") as extracted:
                decompressor = bz2.BZ2Decompressor()
                for chunk in r.iter_content(chunk_size=1024):
                    extracted.write(decompressor.decompress(chunk))

        except Exception:
            self.logger.exception("error downloading ELSA file")
            raise

    def _parse_oval_data(self, path: str, config: dict):
        # normalize and return results
        raw_results = parse(path, config)

        # TODO: remove this filtering once Grype and GrypeDB support the types of version checks necessary
        # to correctly match these.
        # NOTE: Grype and GrypeDB need to identify both a version range AND a specific release version check.
        # See:https://github.com/anchore/anchore-engine/issues/1237 for details and links.
        # This approach is the minimally risk since it only impacts this driver and only ksplice-based packages.
        filterer = KspliceFilterer(logger=self.logger)
        return filterer.filter(raw_results)

    def get(self):
        # download
        self._download()
        return self._parse_oval_data(self.xml_file_path, self.config)


class KspliceFilterer:
    ksplice_regex = re.compile(".*ksplice.*")

    def __init__(self, logger=None):
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    @classmethod
    def _is_ksplice_version(cls, version) -> bool:
        """
        Returns true if the package version (rpm) is a "ksplice" package

        :param version:
        :return:
        """
        epoch, version, release = rpm.split_fullversion(version)
        return cls.ksplice_regex.match(release) is not None

    def filter(self, vuln_dict: dict) -> dict:
        """

        Filters affected packages and ELSAs that are for ksplice packages since the matching logic for these in Grype isn't
        present and it consistently causes false-positives.

        This only removes the FixedIn records, it does not remove the
        parent vuln record. This is necessary to ensure the sync down to engines will get updated properly. No matches will be
        possible with no FixedIn records.

        TODO: this eventually should be removed once support for handling complex version constraints is supported in the feed
        data model and in Grype. See https://github.com/anchore/anchore-engine/issues/1237 for FP description

        :param vuln_dict: dict of vulns where key is distro and version and value is the list of vulns for that version
        :return:
        """
        for version, vuln in vuln_dict.values():  # noqa: B007
            fixes = vuln.get("Vulnerability", {}).get("FixedIn", [])
            if fixes:
                pre_filter_fix_count = len(fixes)

                # sort dictionary by "Name" and "Version" keys
                fixes = sorted(fixes, key=lambda k: (k["Name"], k["Version"]))

                vuln["Vulnerability"]["FixedIn"] = [fix for fix in fixes if not self._is_ksplice_version(fix.get("Version", ""))]

                post_filter_count = len(vuln["Vulnerability"]["FixedIn"])
                filter_diff = pre_filter_fix_count - post_filter_count

                if filter_diff > 0:
                    self.logger.debug(
                        "filtered %s ksplice-based package entries out of total %s package entries for vuln %s",
                        filter_diff,
                        pre_filter_fix_count,
                        vuln["Vulnerability"]["Name"],
                    )

        return vuln_dict



from __future__ import annotations

import logging
from typing import Any


# source: https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
def _addLoggingLevel(levelName: str, levelNum: int, methodName: str | None = None) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    return early if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName) or hasattr(logging, methodName) or hasattr(logging.getLoggerClass(), methodName):
        return

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self: logging.Logger, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message: str, *args: Any, **kwargs: Any) -> None:
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


# note: this needs to be made available to the rest of the application and tests
_addLoggingLevel("TRACE", logging.DEBUG - 5)



from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING, Any

from orjson import loads

from vunnel.utils import archive, http

if TYPE_CHECKING:
    from vunnel.workspace import Workspace


class NVDOverrides:
    __file_name__ = "nvd-overrides.tar.gz"
    __extract_name__ = "nvd-overrides"

    def __init__(  # noqa: PLR0913
        self,
        enabled: bool,
        url: str,
        workspace: Workspace,
        logger: logging.Logger | None = None,
        download_timeout: int = 125,
        retries: int = 5,
    ):
        self.enabled = enabled
        self.__url__ = url
        self.workspace = workspace
        self.download_timeout = download_timeout
        self.retries = retries
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger
        self.__filepaths_by_cve__: dict[str, str] | None = None

    @property
    def url(self) -> str:
        return self.__url__

    def download(self) -> None:
        if not self.enabled:
            self.logger.debug("overrides are not enabled, skipping download...")
            return

        req = http.get(self.__url__, self.logger, stream=True, timeout=self.download_timeout, retries=self.retries)

        file_path = os.path.join(self.workspace.input_path, self.__file_name__)
        with open(file_path, "wb") as fp:
            for chunk in req.iter_content():
                fp.write(chunk)

        archive.extract(file_path, self._extract_path)

    @property
    def _extract_path(self) -> str:
        return os.path.join(self.workspace.input_path, self.__extract_name__)

    def _build_files_by_cve(self) -> dict[str, Any]:
        filepaths_by_cve__: dict[str, str] = {}
        for path in glob.glob(os.path.join(self._extract_path, "**/data/**/", "CVE-*.json"), recursive=True):
            cve_id = os.path.basename(path).removesuffix(".json").upper()
            filepaths_by_cve__[cve_id] = path

        return filepaths_by_cve__

    def cve(self, cve_id: str) -> dict[str, Any] | None:
        if not self.enabled:
            return None

        if self.__filepaths_by_cve__ is None:
            self.__filepaths_by_cve__ = self._build_files_by_cve()

        # TODO: implement in-memory index
        path = self.__filepaths_by_cve__.get(cve_id.upper())
        if path and os.path.exists(path):
            with open(path) as f:
                return loads(f.read())
        return None

    def cves(self) -> list[str]:
        if not self.enabled:
            return []

        if self.__filepaths_by_cve__ is None:
            self.__filepaths_by_cve__ = self._build_files_by_cve()

        return list(self.__filepaths_by_cve__.keys())



from __future__ import annotations

import errno
import logging
import os
import random
import shutil
import time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable


def retry_with_backoff(retries: int = 5, backoff_in_seconds: int = 3) -> Callable[[Any], Any]:
    def rwb(f: Callable[[Any], Any]) -> Callable[[Any], Any]:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            logger = logging.getLogger("utils:retry-with-backoff")
            attempt = 0
            while attempt < retries:
                err = None
                try:
                    return f(*args, **kwargs)
                except KeyboardInterrupt:
                    logger.warning("keyboard interrupt, cancelling request...")
                    raise
                except Exception as e:
                    err = e
                    if attempt >= retries:
                        logger.exception(f"failed after {retries} retries")
                        raise

                # explanation of S311 disable: random number is not used for cryptography
                sleep = backoff_in_seconds * 2**attempt + random.uniform(0, 1)  # noqa: S311
                logger.warning(f"{f} failed with {err}. Retrying in {int(sleep)} seconds (attempt {attempt+1} of {retries})")
                time.sleep(sleep)
                attempt += 1

            raise RuntimeError("max retries reached, failed to execute function")

        return wrapper

    return rwb


def silent_remove(path: str, tree: bool = False) -> None:
    try:
        if tree:
            shutil.rmtree(path)
        else:
            os.remove(path)
    except OSError as e:
        # note: errno.ENOENT = no such file or directory
        if e.errno != errno.ENOENT:
            raise



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .parser import Parser

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=result.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125
    parallelism: int = 4
    full_sync_interval: int = 2
    skip_namespaces: list[str] = field(default_factory=lambda: ["rhel:3", "rhel:4"])


class Provider(provider.Provider):

    __schema__ = schema.OSSchema()
    __distribution_version__ = int(__schema__.major_version)

    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()
        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config

        self.logger.debug(f"config: {config}")

        self.parser = Parser(
            workspace=self.workspace,
            download_timeout=self.config.request_timeout,
            max_workers=self.config.parallelism,
            full_sync_interval=self.config.full_sync_interval,
            skip_namespaces=self.config.skip_namespaces,
            logger=self.logger,
        )

    @classmethod
    def name(cls) -> str:
        return "rhel"

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        with self.results_writer() as writer:
            for namespace, vuln_id, record in self.parser.get(skip_if_exists=self.config.runtime.skip_if_exists):
                namespace = namespace.lower()
                vuln_id = vuln_id.lower()
                writer.write(
                    identifier=os.path.join(namespace, vuln_id),
                    schema=self.__schema__,
                    payload=record,
                )

        return self.parser.urls, len(writer)
